//
// File:        Fortran.java
// Package:     gov.llnl.babel.backend.fortran
// Release:     $Name: release-0-8-0 $
// Revision:    @(#) $Revision: 1.8 $
// Description: Collection of static methods for the FORTRAN binding
//
// Copyright (c) 2000-2002, The Regents of the University of Calfornia.
// Produced at the Lawrence Livermore National Laboratory.
// Written by the Components Team <components@llnl.gov>
// UCRL-CODE-2002-054
// All rights reserved.
//
// This file is part of Babel. For more information, see
// http://www.llnl.gov/CASC/components/. Please read the COPYRIGHT file
// for Our Notice and the LICENSE file for the GNU Lesser General Public
// License.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License (as published by
// the Free Software Foundation) version 2.1 dated February 1999.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
// conditions of the GNU Lesser General Public License for more details.
//
// You should have recieved a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


package gov.llnl.babel.backend.fortran;

import gov.llnl.babel.BabelConfiguration;
import gov.llnl.babel.backend.CodeConstants;
import gov.llnl.babel.backend.CodeGenerationException;
import gov.llnl.babel.backend.mangler.NameMangler;
import gov.llnl.babel.symbols.Method;
import gov.llnl.babel.symbols.SymbolID;
import gov.llnl.babel.symbols.Type;

/**
 * Provide a collection of static methods to provide the mapping of SIDL
 * concepts into FORTRAN.  This class provides the mapping of symbol names
 * to FORTRAN symbols, the mapping of the wrapper code to a set of files,
 * and the mapping of types.
 */
public class Fortran implements CodeConstants 
{

  /**
   * The order of all of the following arrays must match the ordering of 
   * the constants in the Type class.  A separate version is maintained to
   * facilitate translation for each version of fortran.
   *
   * @see gov.llnl.babel.symbols.Type
   */
  private final static String s_f77_types[] = {
    "void",                     // void
    "logical",                  // bool
    "character",                // character
    "double complex",           // dcomplex
    "double precision",         // double
    "complex",                  // fcomplex
    "real",                     // float
    "integer*4",                // int
    "integer*8",                // long
    "integer*8",                // opaque
    "character*(*)",            // string
    "integer",                  // enum
    "integer*8",                // interface
    "integer*8",                // class
    "integer*8",                // symbol
    "integer*8"                 // array
  };

  private final static String s_f90_types[] = {
    "void",  					// void
    "logical",  				// bool
    "character (len=1)",  			// character
    "complex (selected_real_kind(15, 307))",	// dcomplex
    "real (selected_real_kind(15, 307))",	// double
    "complex (selected_real_kind(6, 37))",	// fcomplex
    "real (selected_real_kind(6, 37))",		// float
    "integer (selected_int_kind(9))",		// int
    "integer (selected_int_kind(18))",		// long
    "integer (selected_int_kind(18))",		// opaque
    "character*(*)",           			// string
    "integer (selected_int_kind(9))",		// enum
    "integer (selected_int_kind(18))",		// interface
    "integer (selected_int_kind(18))",		// class
    "integer (selected_int_kind(18))",		// symbol
    "integer (selected_int_kind(18))"		// array
  };
  
  /**
   * This array stores the name of the C type that corresponds to the FORTRAN
   * type for a particular SIDL type.  For example, an opaque maps into
   * <code>integer*8</code> in FORTRAN.  When the <code>integer*8</code>
   * comes into C, it is a <code>int64_t</code>.
   */
  private final static String[] s_cTypeMatchingFortran = {
    "void",                     // void
    "SIDL_F77_Bool",            // bool
    "char",                     // character
    "struct SIDL_dcomplex",     // dcomplex
    "double",                   // double
    "struct SIDL_fcomplex",     // fcomplex
    "float",                    // float
    "int32_t",                  // int
    "int64_t",                  // long
    "int64_t",                  // opaque
    "char *",                   // string
    "int",                      // enum
    "int64_t",                  // interface
    "int64_t",                  // class
    "int64_t",                  // symbol
    "int64_t"                   // array
  };
  private final static int    s_Fortran_Bool = 1;
  private final static String s_F90_Bool = "SIDL_F90_Bool";

  /**
   * Indicator of the version of FORTRAN being generated.  True means
   * F90; otherwise, F77 assumed.
   */
  private final static boolean s_isF90 = 
      BabelConfiguration.getInstance().getTargetLanguage().equals("f90");


  /**
   * This class is not intended for public instantiation.
   */
  private Fortran()
  {
    // not intended for instantiation
  }

  /**
   * Convert a symbol name into string with the pieces of the symbol joined
   * together with underline characters.
   *
   * @param id  the symbol id to convert.
   * @return a string representation of the symbol with periods replaced
   * with underline characters.
   */
  public static String getSymbolName(SymbolID id) {
    return id.getFullName().replace('.', '_');
  }

  /**
   * Return the name of the stub file for a particular symbol.
   * The stub file is a C file that receives calls from FORTRAN client.
   *
   * @param id  the symbol whose stub file will be returned
   * @return the filename of a C file containing the implementation of
   *         the FORTRAN stubs.
   */
  public static String getStubFile(SymbolID id) {
    return getSymbolName(id) + "_fStub.c";
  }

  public static String getStubNameFile(SymbolID id) {
    return getSymbolName(id) + "_fAbbrev.h";
  }

  /**
   * Return the name of the F77 stub documentation file for a particular
   * symbol. The documentation file is FORTRAN pseudo-code to
   * document the calling interface for FORTRAN clients.
   *
   * @param id  the symbol whose stub documentation file will be returned
   * @return the filename of a text file containing the documentation
   *         of the FORTRAN stubs.
   */
  public static String getStubDocFile(SymbolID id) {
    return getSymbolName(id) + ".fif";
  }

  /**
   * Return the name of the F90 stub module file for a particular
   * symbol. The module file is an F90 file used to package "global
   * data, derived types and their associated operations, interface
   * blocks, and namelist groups."
   *
   * @param id  the symbol whose stub documentation file will be returned
   * @return the filename of a text file containing the documentation
   *         of the FORTRAN stubs.
   */
  public static String getModuleFile(SymbolID id) {
    return getSymbolName(id) + "_Mod." + CodeConstants.C_F90_IMPL_EXTENSION;
  }

  /**
   * Return the name of the stub file for a particular enumeration.
   * The stub file is a DEC FORTRAN include file that defines
   * integer parameters that hold the values of the enumerations.
   *
   * @param id  the symbol whose stub file will be returned
   * @return the filename of a FORTRAN inc file containing the implementation of
   *         the FORTRAN stubs.
   */
  public static String getEnumStubFile(SymbolID id) {
    return getSymbolName(id) + ".inc";
  }

  public static String getEnsureArray(Type arrayType) {
    switch(arrayType.getDetailedType()) {
    case Type.INTERFACE:
    case Type.CLASS:
      return "SIDL_interface__array_ensure((struct SIDL_interface__array *)";
    case Type.ENUM:
      return "SIDL_int__array_ensure((struct SIDL_int__array *)";
    default:
      return "SIDL_" + arrayType.getTypeString() + "__array_ensure(";
    }
  }

  public static String getDelRefArray(Type arrayType) {
    switch(arrayType.getDetailedType()) {
    case Type.INTERFACE:
    case Type.CLASS:
      return "SIDL_interface__array_deleteRef((struct SIDL_interface__array *)";
    case Type.ENUM:
      return "SIDL_int__array_deleteRef((struct SIDL_int__array *)";
    default:
      return "SIDL_" + arrayType.getTypeString() +
        "__array_deleteRef((struct SIDL_" +
        arrayType.getTypeString() + "__array *)";
    }
  }

  /**
   * Return the name of the stub impl file for a particular enumeration.
   * The stub file is a C source file that holds the array of enumeration
   * stubs.
   *
   * @param id  the symbol whose stub file will be returned
   * @return the filename of a C file containing the implementation of
   *         the FORTRAN enum array stubs.
   */
  public static String getEnumStubImpl(SymbolID id) {
    return getSymbolName(id) + "_fStub.c";
  }

  /**
   * Return the name of the file that holds the implementation of the
   * skeletons for FORTRAN. The skeleton file is written in C, and it is
   * the glue between the IOR and a FORTRAN.
   */
  public static String getSkelFile(SymbolID id) {
    return getSymbolName(id) + "_fSkel.c";
  }

  /**
   * Return the appropriate FORTRAN extension.
   *
   * @return the FORTRAN file extension.
   */
  public static String getImplExtension() {
    if (s_isF90) {
      return CodeConstants.C_F90_IMPL_EXTENSION;
    } else {
      return CodeConstants.C_F77_IMPL_EXTENSION;
    }
  }

  /**
   * Return the name of the file that hold the implementation of the
   * FORTRAN SIDL object. The BABEL system generates the subroutine
   * declarations, but the user needs to fill in the subroutine bodies.
   *
   * @param id the name of the symbol
   * @return the filename for the SIDL object implementation.
   */
  public static String getImplFile(SymbolID id) {
    return getSymbolName(id) + "_Impl." + getImplExtension();
  }

  /**
   * Get the function name that should be used for the FORTRAN stubs to the
   * SIDL object methods.  This method does not take into account the
   * compiler specific issues (i.e. whether the symbol should be all upper
   * or lower case or whether it has underscores appended); it provides
   * the starting name with potentially mixed case.
   *
   * @param  id     the name of the symbol who has the method.
   * @param  method information about the method to be named.
   * @return the name of the function to be used in the FORTRAN stub.
   */
  public static String getMethodStubName(SymbolID id, Method method) {
    return getSymbolName(id) + "_" + method.getLongMethodName() 
           + getMethodSuffix();
  }

  /**
   * Return the appropriate version of Fortran (i.e., 77 or 90).
   *
   * @return the version of Fortran being generated
   */
  public static int getFortranVersion() {
    int fversion;
    if (s_isF90) {
      fversion = CodeConstants.C_F90_VERSION;
    } else {
      fversion = CodeConstants.C_F77_VERSION;
    }
    return fversion;
  }

  /**
   * Return the appropriate FORTRAN method name suffix based on the version.
   *
   * @return the FORTRAN method name suffix.
   */
  public static String getMethodSuffix() {
    if (s_isF90) {
      return CodeConstants.C_F90_METHOD_SUFFIX;
    } else {
      return CodeConstants.C_F77_METHOD_SUFFIX;
    }
  }

  /**
   * Return the appropriate FORTRAN impl method name suffix based.
   *
   * @return the FORTRAN impl method name suffix.
   */
  public static String getImplMethodSuffix() {
    if (s_isF90) {
      return CodeConstants.C_F90_IMPL_METHOD_SUFFIX;
    } else {
      return CodeConstants.C_F77_IMPL_METHOD_SUFFIX;
    }
  }

  /**
   * Return the appropriate version of SIDLFortran##Symbol.
   *
   * @return the appropriate version of SIDLFortran##Symbol.
   */
  public static String getFortranSymbol() {
    return "SIDLFortran" + getFortranVersion() + "Symbol";
  }

  /**
   * Generate the name of the array destructor function.
   */
  public static String getArrayDestructor(SymbolID id) {
    return id.getFullName().replace('.', '_') + "__array_deleteRef"
           + getMethodSuffix();
  }

  /**
   * Generate the name of the array constructor function.
   */
  public static String getArrayConstructor(SymbolID id) {
    return id.getFullName().replace('.', '_') + "__array_createRow" 
           + getMethodSuffix();
  }

  /**
   * Generate the name of the array set element function.
   */
  public static String getArraySet(SymbolID id) {
    return id.getFullName().replace('.', '_') + "__array_set" 
           + getMethodSuffix();
  }

  /**
   * Generate the name of the array set element function.
   */
  public static String getArraySet(SymbolID id, int numArgs) {
    return id.getFullName().replace('.', '_') + "__array_set" +
      Integer.toString(numArgs) + getMethodSuffix();
  }

  /**
   * Generate the name of the array get element function.
   */
  public static String getArrayGet(SymbolID id) {
    return id.getFullName().replace('.', '_') + "__array_get"
           + getMethodSuffix();
  }

  /**
   * Generate the name of the array get element function.
   */
  public static String getArrayGet(SymbolID id, int numArgs) {
    return id.getFullName().replace('.', '_') + "__array_get" +
      Integer.toString(numArgs) + getMethodSuffix();
  }

  /**
   * Generate the name of the array dimension access function.
   */
  public static String getArrayDimen(SymbolID id) {
    return id.getFullName().replace('.', '_') + "__array_dimen" 
           + getMethodSuffix();
  }

  /**
   * Generate the name of the array lower bound access function.
   */
  public static String getArrayLower(SymbolID id) {
    return id.getFullName().replace('.', '_') + "__array_lower"
           + getMethodSuffix();
  }

  /**
   * Generate the name of the array upper bound access function.
   */
  public static String getArrayUpper(SymbolID id) {
    return id.getFullName().replace('.', '_') + "__array_upper"
           + getMethodSuffix();
  }

  /**
   * Return the function name that should be used for the FORTRAN subroutine
   * that implements a particular method.  The returned string leaves
   * the cast of the components of the name unchanged.  If <code>id</code>
   * and the method name have mixed case, this method returns a mixed case
   * string.
   *
   * @param  id     the name of the symbol who has the method.
   * @param  method the information about the method.
   * @param  mang   the name mangler being used to address long names.
   * @return the name of the FORTRAN subroutine that implements this method.
   * @exception gov.llnl.babel.backend.CodeGenerationException
   *    the name mangler is broken.
   */
  public static String getMethodImplName(SymbolID id,
                                         Method method,
                                         NameMangler mang)
    throws CodeGenerationException
  {
    try {
      return mang.shortName(id.getFullName(), 
                            method.getLongMethodName(),
                            getImplMethodSuffix());
    }
    catch (java.io.UnsupportedEncodingException uee) {
      throw new CodeGenerationException(uee.getMessage());
    }
  }

  /**
   * Return the C function name that should be used for the skeleton
   * method.  This is a C function that is called by the IOR.
   *
   * @param  id     the name of the symbol who has the method
   * @param  method information about the method.
   * @return the name of the C skeleton function.
   */
  public static String getMethodSkelName(SymbolID id, Method method) {
    return "skel_" + method.getLongMethodName();
  }

  /**
   * Return the C type corresponding to the FORTRAN type corresponding to
   * a particular SIDL type.  This is the type that a C subroutine would
   * need to pass to FORTRAN, or the type that a C subroutine could expect
   * to receive from a FORTRAN caller.
   *
   * @parm type  the SIDL type description.
   * @return the C type corresponding to the FORTRAN type corresponding
   *         to the SIDL type.
   * @exception gov.llnl.babel.backend.CodeGenerationException
   *    the type is unsupported.
   */
  public static String getFortranTypeInC(Type type)
    throws CodeGenerationException
  {
    final int t = type.getDetailedType();
    if (t >= 0 && t < s_cTypeMatchingFortran.length) {
      if (t == s_Fortran_Bool && s_isF90) {
        return s_F90_Bool;
      } else {
        return s_cTypeMatchingFortran[t];
      }
    }
    throw new CodeGenerationException("Unknown supported " + t);
  }

  /**
   * Return the SIDL Fortran prefix for macros and types.
   *
   * @return the SIDL Fortran prefix
   */
  public static String getFortranPrefix() {
    return "SIDL_F" + getFortranVersion();
  }

  /**
   * Return the FORTRAN type declaration corresponding to <code>type</code>,
   * a particular SIDL type.
   *
   * @param  type  a SIDL type description.
   * @return the FORTRAN type used to some something of the given SIDL
   * type.
   * @exception gov.llnl.babel.backend.CodeGenerationException
   *    the type is unsupported.
   */
  public static String getReturnString(Type type)
    throws CodeGenerationException
  {
    final int t = type.getDetailedType();
    if (s_isF90) {
      if (t >= 0 && t < s_f90_types.length) {
        return s_f90_types[t];
      }
    } else {
      if (t >= 0 && t < s_f77_types.length) {
        return s_f77_types[t];
      }
    }
    throw new CodeGenerationException("Unknown supported " + t);
  }
  /**
   * Return the FORTRAN type declaration corresponding to the <code>int</code>
   * representation of a particular SIDL type.
   *
   * @param  type  a SIDL type (from Type)
   * @return the FORTRAN type used to some something of the given SIDL
   * type.
   * @exception gov.llnl.babel.backend.CodeGenerationException
   *    the type is unsupported.
   */
  public static String getReturnString(int type)
    throws CodeGenerationException
  {
    if (s_isF90) {
      if (type >= 0 && type < s_f90_types.length) {
        return s_f90_types[type];
      }
    } else {
      if (type >= 0 && type < s_f77_types.length) {
        return s_f77_types[type];
      }
    }
    throw new CodeGenerationException("Unknown supported " + type);
  }
}
