//
// File:        GenerateClient.java
// Package:     gov.llnl.babel.backend.fortran
// Release:     $Name: release-0-8-0 $
// Revision:    @(#) $Revision: 1.4 $
// Description: generate C code to allow FORTRAN to access BABEL
//
// Copyright (c) 2000-2002, The Regents of the University of Calfornia.
// Produced at the Lawrence Livermore National Laboratory.
// Written by the Components Team <components@llnl.gov>
// UCRL-CODE-2002-054
// All rights reserved.
//
// This file is part of Babel. For more information, see
// http://www.llnl.gov/CASC/components/. Please read the COPYRIGHT file
// for Our Notice and the LICENSE file for the GNU Lesser General Public
// License.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License (as published by
// the Free Software Foundation) version 2.1 dated February 1999.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
// conditions of the GNU Lesser General Public License for more details.
//
// You should have recieved a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


package gov.llnl.babel.backend.fortran;

import gov.llnl.babel.BabelConfiguration;
import gov.llnl.babel.backend.CodeGenerationException;
import gov.llnl.babel.backend.CodeGenerationFactory;
import gov.llnl.babel.backend.CodeGenerator;
import gov.llnl.babel.backend.FileManager;
import gov.llnl.babel.backend.fortran.ModuleSource;
import gov.llnl.babel.backend.fortran.StubDoc;
import gov.llnl.babel.backend.fortran.StubSource;
import gov.llnl.babel.backend.writers.LanguageWriterForC;
import gov.llnl.babel.backend.writers.LanguageWriterForFortran;
import gov.llnl.babel.symbols.Symbol;
import gov.llnl.babel.symbols.SymbolID;
import gov.llnl.babel.symbols.SymbolTable;
import java.io.PrintWriter;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;


/**
 * This class implements the {@link gov.llnl.babel.backend.CodeGenerator
 * CodeGenerator} interface for the FORTRAN client side code generator.
 * This generator creates C source code to provide the glue between
 * FORTRAN clients and the IOR.
 */
public class GenerateClient implements CodeGenerator 
{
  /**
   * Create a new instance.
   */
  public GenerateClient() {
    
  }

  /**
   * If <code>symbol</code> is a class or interface, create a stub file for
   * it.
   *
   * @param symbol  a particlar symbol.
   * @exception gov.llnl.babel.backend.CodeGenerationException
   *  a catch all exception.
   */
  private void generateStubSource(Symbol symbol)
    throws CodeGenerationException 
  {
    PrintWriter pw = null, dw = null, nw = null, mw = null;
    LanguageWriterForFortran forWriter;
    final int type = symbol.getSymbolType();
    boolean isF90 = "f90".
      equals(BabelConfiguration.getInstance().getTargetLanguage());
    switch (type){
    case Symbol.CLASS:
    case Symbol.INTERFACE:
      try {
        String f = Fortran.getStubFile(symbol.getSymbolID());
        pw       = FileManager.getInstance().createFile(symbol, "STUBSRCS", f);

        LanguageWriterForC writer = new LanguageWriterForC(pw);
        StubSource.generateCode(symbol, writer);

        if (isF90) {
          String nf = Fortran.getStubNameFile(symbol.getSymbolID());
          nw        = FileManager.getInstance().createFile(symbol, "STUBHDRS", 
                                                           nf);
          writer    = new LanguageWriterForC(nw);
          AbbrevHeader.generateCode(symbol, writer);

          String mf = Fortran.getModuleFile(symbol.getSymbolID());
          mw        = FileManager.getInstance().createFile(symbol, 
                                                           "STUBMODULESRCS", 
                                                           mf);
	  forWriter = new LanguageWriterForFortran(mw);
          ModuleSource.generateCode(symbol, forWriter);
        } else {
          String df = Fortran.getStubDocFile(symbol.getSymbolID());
          dw        = FileManager.getInstance().createFile(symbol, "STUBDOCS", 
                                                           df);
	  forWriter = new LanguageWriterForFortran(dw);
          StubDoc.generateCode(symbol, forWriter);
        }
      }
      finally {
        if (pw != null) {
          pw.close();
        }
        if (dw != null) {
          dw.close();
        }
        if (nw != null) {
          nw.close();
        }
        if (mw != null) {
          mw.close();
        }
      }
      break;
    case Symbol.ENUM:
// To Do:  Should a module be created for this as well? 
      try {
        String f  = Fortran.getEnumStubFile(symbol.getSymbolID());
        String nf = Fortran.getStubNameFile(symbol.getSymbolID());
        pw        = FileManager.getInstance().
          createFile(symbol, "STUBFORTRANINC", f);
        LanguageWriterForFortran writer = new LanguageWriterForFortran(pw);
        StubSource.generateCode(symbol, writer);
        writer.close();
        f = Fortran.getEnumStubImpl(symbol.getSymbolID());
        pw = FileManager.getInstance().createFile(symbol, "STUBSRCS", f);
        LanguageWriterForC clw = new LanguageWriterForC(pw);
        StubSource.generateCode(symbol, clw);
        if (isF90) {
          nw  = FileManager.getInstance().createFile(symbol, "STUBHDRS", nf);
          clw = new LanguageWriterForC(nw);
          AbbrevHeader.generateCode(symbol, clw);
        }
      }
      finally {
        if (pw != null) {
          pw.close();
        }
        if (nw != null) {
          nw.close();
        }
      }
      break;
    }
  }
  
  /**
   * Given a set of symbol ids, this method will generate FORTRAN stubs
   * (written in C) for all the symbols in the set as needed.  Some symbol
   * types may not require a stub file, for example packages, and these are 
   * silently skipped.
   *
   * @param symbols    a set of symbol id (symbol names) for whom stubs
   *                   should be written as needed.  Each object in the
   *                   set should be a {@link
   *                   gov.llnl.babel.symbols.SymbolID SymbolID}.
   * @exception gov.llnl.babel.backend.CodeGenerationException
   *   a catch all exception to indicate problems in the code generation
   *   phase.
   */
  public void generateCode(Set symbols)
    throws CodeGenerationException 
  {
    Iterator i = symbols.iterator();
    SymbolTable table = SymbolTable.getInstance();
    while (i.hasNext()) {
      SymbolID id     = (SymbolID)i.next();
      Symbol   symbol = table.lookupSymbol(id);
      if (symbol != null) {
        generateStubSource(symbol);
      }
    }
  }

  public String getType()
  {
    return "stub";
  }

  public boolean getUserSymbolsOnly()
  {
    return false;
  }

  public Set getLanguages()
  {
    Set result = new TreeSet();
    result.add("f77");
    result.add("f90");
    return result;
  }
}
