//
// File:        ModuleSource.java
// Package:     gov.llnl.babel.backend.fortran
// Release:     $Name: release-0-8-0 $
// Revision:    @(#) $Revision: 1.3 $
// Description: Generate a module file for FORTRAN 90 clients
//
// Copyright (c) 2002, The Regents of the University of Calfornia.
// Produced at the Lawrence Livermore National Laboratory.
// Written by the Components Team <components@llnl.gov>
// UCRL-CODE-2002-054
// All rights reserved.
//
// This file is part of Babel. For more information, see
// http://www.llnl.gov/CASC/components/. Please read the COPYRIGHT file
// for Our Notice and the LICENSE file for the GNU Lesser General Public
// License.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License (as published by
// the Free Software Foundation) version 2.1 dated February 1999.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
// conditions of the GNU Lesser General Public License for more details.
//
// You should have recieved a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


package gov.llnl.babel.backend.fortran;

import gov.llnl.babel.BabelConfiguration;
import gov.llnl.babel.backend.CodeConstants;
import gov.llnl.babel.backend.CodeGenerationException;
import gov.llnl.babel.backend.IOR;
import gov.llnl.babel.backend.fortran.Fortran;
import gov.llnl.babel.backend.fortran.StubSource;
import gov.llnl.babel.backend.writers.LanguageWriterForFortran;
import gov.llnl.babel.symbols.Argument;
import gov.llnl.babel.symbols.Comment;
import gov.llnl.babel.symbols.Extendable;
import gov.llnl.babel.symbols.Method;
import gov.llnl.babel.symbols.Symbol;
import gov.llnl.babel.symbols.SymbolID;
import gov.llnl.babel.symbols.SymbolTable;
import gov.llnl.babel.symbols.Type;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

/**
 * This class provides the ability to write a FORTRAN 90 module for
 * a SIDL class/interface. 
 */
public class ModuleSource {

  /**
   * Indicator of version of FORTRAN being generated.
   */
  private final static boolean d_isF90 = "f90".
    equals(BabelConfiguration.getInstance().getTargetLanguage());

  /**
   * This is the output device.
   */
  private LanguageWriterForFortran d_lw;


  /**
   * Generate an instance to write the module for a FORTRAN 90
   * client.
   * 
   * @param writer  the output device to which the FORTRAN 90 module
   *                  should be written.
   */
  public ModuleSource(LanguageWriterForFortran writer) {
    d_lw = writer;
  }


  /**
   * Generate the argument list for the subroutine, including the
   * opening and closing parens.
   *
   * @param List     List of arguments to be printed
   */
  private void printArgs(List args)
  {
    d_lw.print("(");
    Iterator i = args.iterator();
    while (i.hasNext()) {
      d_lw.print(((Argument)i.next()).getFormalName());
      if (i.hasNext()) {
        d_lw.print(", ");
      }
    }
    d_lw.println(")");
  }


  /**
   * Write the FORTRAN 90 module for a subroutine that corresponds to a 
   * SIDL class/interface method. This writes method signatures and declares 
   * the types of the arguments.
   * 
   * @param m     the method whose module method is to be written.
   * @param id    the name of the class that owns this method.
   * @exception   gov.llnl.babel.backend.CodeGenerationException
   *   a catch all exception to indicate problems in the code generation
   *   phase.
   */
  private void extendAndGenerate(Method   m,
                                 SymbolID id)
    throws CodeGenerationException
  {
    List extendedArgs = StubSource.extendArgs(id, m);
    d_lw.writeComment(m, false);
    d_lw.print("subroutine " + m.getLongMethodName());
    printArgs(extendedArgs);

    d_lw.increaseTabLevel();
    d_lw.println("implicit none");

    Iterator i = extendedArgs.iterator();
    while (i.hasNext()) {
      Argument a = (Argument)i.next();
      d_lw.writeCommentLine(a.getArgumentString());
      d_lw.println(Fortran.getReturnString(a.getType()) + " :: " 
                   + a.getFormalName());
    }
    d_lw.println();
    d_lw.print("call " + Fortran.getMethodStubName(id, m));
    printArgs(extendedArgs);

    d_lw.decreaseTabLevel();
    d_lw.println();
    d_lw.println("end subroutine " + m.getLongMethodName());
  }


  /**
   * Generate a create method object.
   */
  private Method create(SymbolID id)
  {
    Method m = new Method();
    m.setMethodName("_create");
    m.setDefinitionModifier(Method.STATIC);
    String[] s = new String[1];
    s[0] = "Create an instance of class " + id.getFullName();
    m.setComment(new Comment(s));
    m.setReturnType(new Type(id));
    return m;
  }


  /**
   * Add the implicit stub methods to the list of those that must
   * be included.
   *
   * @param ext the class whose module file is being written.
   */
  private Collection extendMethods(Extendable ext) {
    Collection allMethods = ext.getMethods(true);
    final SymbolID id = ext.getSymbolID();
    ArrayList  extendedMethods = new ArrayList(allMethods.size()+2);
    if (!ext.isAbstract()) {
      extendedMethods.add(create(id));
    }
    extendedMethods.add
      (IOR.getBuiltinMethod(IOR.CAST, id, ext.isInterface()));

    extendedMethods.addAll(allMethods);
    return extendedMethods;
  }


  /**
   * Write #include for all the abbreviation files for
   * referenced symbols.
   */
  private void writeIncludes(Extendable ext)
  {
    Iterator i = ext.getSymbolReferences().iterator();
    d_lw.disableLineBreak();
    while (i.hasNext()) {
      SymbolID id = (SymbolID)i.next();
      d_lw.printlnUnformatted("#include \"" +
                              Fortran.getStubNameFile(id) + "\"");
    }
    d_lw.enableLineBreak();
  }


  /**
   * Generate the FORTRAN 90 module file for a SIDL class.  
   * 
   * @param ext    the SIDL class whose module is to be written.
   * @exception    gov.llnl.babel.backend.CodeGenerationException
   *   a catch all exception to indicate problems in the code generation
   *   phase.
   */
  public void generateCode(Extendable ext)
    throws CodeGenerationException
  {
    Collection methods = extendMethods(ext);
    final SymbolID id = ext.getSymbolID();
    final String name = ext.getFullName().replace('.', '_');

    d_lw.writeBanner(ext, Fortran.getModuleFile(id), false,
                     CodeConstants.C_FORTRAN_MODULE_PREFIX 
                     + id.getFullName());
    d_lw.println();
    d_lw.writeComment(ext, false);
    d_lw.println();
    writeIncludes(ext);
    d_lw.println();
    d_lw.println("module " + name);

    d_lw.println();
    d_lw.println("contains");
    d_lw.println();

    d_lw.increaseTabLevel();
    Iterator i = methods.iterator();
    while (i.hasNext()) {
      d_lw.println();
      d_lw.println();

      extendAndGenerate((Method)i.next(), id);
    }
    d_lw.decreaseTabLevel();
    d_lw.println();
    d_lw.println("end module " + name);
  }


  /**
   * Generate the FORTRAN 90 module file for a SIDL class.  
   *
   * Note:  This is the assumed entry point; otherwise, the test for
   * the version of the language should be repeated.
   *
   * @exception gov.llnl.babel.backend.CodeGenerationException
   *   a catch all exception to indicate problems in the code generation
   *   phase.
  */
  public static void generateCode(Symbol                   sym,
                                  LanguageWriterForFortran writer)
    throws CodeGenerationException
  {
    if (d_isF90) {
      if (sym instanceof Extendable) {
        ModuleSource modFile = new ModuleSource(writer);
        modFile.generateCode((Extendable)sym);
      }
    } else {  // Assuming this means F77
      throw new CodeGenerationException("Generation of module files only " +
                                        "supported for FORTRAN 90");
    }
  }
}
